//	Kameratimer 1.0
//
//	Copyright (C) 2009 Paul Wilhelm <paul@mosfetkiller.de>
//	http://mosfetkiller.de/?s=kameratimer
//
//	Modul: main.c | Hauptprogramm

#include "main.h"

//	Main
int main(void)
{
//	_delay_ms(100);

	//	Initialisierung
	init();

	unsigned char brightness_read = eeprom_read_byte(BRIGHTNESS_EEPROM_ADDR_A);
	unsigned char pulse_length_read[3] = { eeprom_read_byte(PULSE_LENGTH_EEPROM_ADDR_A), eeprom_read_byte(PULSE_LENGTH_EEPROM_ADDR_B), eeprom_read_byte(PULSE_LENGTH_EEPROM_ADDR_C) };
	unsigned char shots_total_read = eeprom_read_byte(SHOTS_TOTAL_EEPROM_ADDR_A);

	//	Unsinnige EEPROM-Werte initialisieren (bei erstmaligem Start der Software)
	if (brightness_read > 10) { brightness_read = 10; eeprom_write_byte(BRIGHTNESS_EEPROM_ADDR_A, brightness_read); }
	if (pulse_length_read[0] > 99) { pulse_length_read[0] = 0; eeprom_write_byte(PULSE_LENGTH_EEPROM_ADDR_A, pulse_length_read[0]); }
	if (pulse_length_read[1] > 99) { pulse_length_read[1] = 0; eeprom_write_byte(PULSE_LENGTH_EEPROM_ADDR_B, pulse_length_read[1]); }
	if (pulse_length_read[2] > 99) { pulse_length_read[2] = 0; eeprom_write_byte(PULSE_LENGTH_EEPROM_ADDR_C, pulse_length_read[2]); }
	if (shots_total_read > 99) { shots_total_read = 0; eeprom_write_byte(SHOTS_TOTAL_EEPROM_ADDR_A, shots_total_read); }

	brightness = brightness_read;
	for (unsigned char i = 0; i < 3; i++) pulse_length[i] = pulse_length_read[i];
	shots_total = shots_total_read;

	//	Hauptschleife
	while (1)
	{

		//	Timer Setup
		if (btn_press(BTN_SETUP, FALSE))
		{
			//	GGf. geänderte Einstellungen im EEPROM speichern (falls man gerade aus dem Parameter- bzw. Impulsmenü kommt)
			if (brightness_read != brightness) { eeprom_write_byte(BRIGHTNESS_EEPROM_ADDR_A, brightness); brightness_read = brightness; }
			if (pulse_length_read[0] != pulse_length[0]) { eeprom_write_byte(PULSE_LENGTH_EEPROM_ADDR_A, pulse_length[0]); pulse_length_read[0] = pulse_length[0]; }
			if (pulse_length_read[1] != pulse_length[1]) { eeprom_write_byte(PULSE_LENGTH_EEPROM_ADDR_B, pulse_length[1]); pulse_length_read[1] = pulse_length[1]; }
			if (pulse_length_read[2] != pulse_length[2]) { eeprom_write_byte(PULSE_LENGTH_EEPROM_ADDR_C, pulse_length[2]); pulse_length_read[2] = pulse_length[2]; }
			if (shots_total_read != shots_total) { eeprom_write_byte(SHOTS_TOTAL_EEPROM_ADDR_A, shots_total); shots_total_read = shots_total; }

			//	Blinkende Punkte aus
			blink = FALSE;

			//	Die Zeit wird gerade mindestens erstmalig gesetzt
			timer_set = TRUE;

			//	Laufender Timer stoppt
			timer_running = FALSE;

			//	Relais und Punkt auf jeden Fall aus
			relais = FALSE;
			dot[0] = 0;
			clear_bit(RELAIS_port, RELAIS_bit);

			//	Schusszähler resetten
			shots_total_this = 0;

			//	Stunden, Minuten, Sekunden durch zwei Punkte trennen
			dot[2] = 1;
			dot[4] = 1;

			while (btn_press(BTN_SETUP, FALSE))
			{
				//	Timereinstellung anzeigen
				display_number(timer_max[SECONDS], 0);
				display_number(timer_max[MINUTES], 1);
				display_number(timer_max[HOURS], 2);

				//	Stunden einstellen
				check_btn_varchange(BTN_HOURS_UP, &timer_max[HOURS], 0, 99, TRUE, 2, FALSE);
				check_btn_varchange(BTN_HOURS_DN, &timer_max[HOURS], 0, 99, FALSE, 2, FALSE);

				//	Minuten einstellen
				check_btn_varchange(BTN_MINUTES_UP, &timer_max[MINUTES], 0, 59, TRUE, 1, FALSE);
				check_btn_varchange(BTN_MINUTES_DN, &timer_max[MINUTES], 0, 59, FALSE, 1, FALSE);

				//	Sekunden einstellen
				check_btn_varchange(BTN_SECONDS_UP, &timer_max[SECONDS], 0, 59, TRUE, 0, FALSE);
				check_btn_varchange(BTN_SECONDS_DN, &timer_max[SECONDS], 0, 59, FALSE, 0, FALSE);
			}
		}

		//	Start-Stellung
		if (btn_press(BTN_START, FALSE))
		{

			//	Noch keine Zeit eingestellt: Parameter-Setup
			if (!timer_set)
			{
				//	Stunden, Minuten, Sekunden durch zwei Punkte trennen
				dot[2] = 1;
				dot[4] = 1;

				//	Blinkende Punkte ein
				if (!blink) blink = TRUE;

				//	Werte anzeigen
				display_number(brightness, 2);
				display[2] = -1; display[3] = -1;	//	Mittlere Doppelstelle aus
				display_number(shots_total, 0);


				//	Helligkeit einstellen
				check_btn_varchange(BTN_HOURS_UP, &brightness, 1, 10, TRUE, 2, FALSE);
				check_btn_varchange(BTN_HOURS_DN, &brightness, 1, 10, FALSE, 2, FALSE);

				//	Anzahl der Auslösungen einstellen
				check_btn_varchange(BTN_SECONDS_UP, &shots_total, 0, 99, TRUE, 0, FALSE);
				check_btn_varchange(BTN_SECONDS_DN, &shots_total, 0, 99, FALSE, 0, FALSE);


				//	Extra-Menü: Einstellung der Impulslänge
				if (btn_press(BTN_MINUTES_UP, TRUE) || btn_press(BTN_MINUTES_DN, TRUE))
				{
					//	Werte auf dem Display darstellen
					display_number(pulse_length[2], 2);
					display_number(pulse_length[1], 1);
					display_number(pulse_length[0], 0);

					//	Punkte blinken nicht
					blink = FALSE;

					//	Stunden, Minuten, Sekunden durch zwei Punkte trennen
					dot[2] = 1;
					dot[4] = 1;

					//	Ziffern blinken lassen
					for (unsigned char i = 0; i < 3; i++) blink_digit[i] = TRUE;

					//	Idle-Timer
					my_timer_2 = 0;

					//	Millisekundenzähler resetten, damit das Ziffernblinken gleichmäßig startet
					mseconds = 0;

					//	Warten, bis Minutentaster wieder losgelassen wurde
					while ((button[BTN_MINUTES_UP] + button[BTN_MINUTES_DN]) > 0);

					//	Bei abgelaufenem Idle-Timer oder gedrückten Setup-Schalter Menü beenden
					while (my_timer_2 < PULSE_LENGTH_MENU_IDLE_EXIT && !btn_press(BTN_SETUP, FALSE))
					{
						//	Bei jedem Tastendruck den Idle-Timer resetten

						//	Impulsdauer-Minuten einstellen
						if (check_btn_varchange(BTN_HOURS_UP, &pulse_length[2], 0, 99, TRUE, 2, TRUE)) my_timer_2 = 0;
						if (check_btn_varchange(BTN_HOURS_DN, &pulse_length[2], 0, 99, FALSE, 2, TRUE)) my_timer_2 = 0;

						//	Impulsdauer-Sekunden einstellen
						if (check_btn_varchange(BTN_MINUTES_UP, &pulse_length[1], 0, 59, TRUE, 1, TRUE)) my_timer_2 = 0;
						if (check_btn_varchange(BTN_MINUTES_DN, &pulse_length[1], 0, 59, FALSE, 1, TRUE)) my_timer_2 = 0;

						//	Impulsdauer-Zehntelsekunden einstellen
						if (check_btn_varchange(BTN_SECONDS_UP, &pulse_length[0], 0, 99, TRUE, 0, TRUE)) my_timer_2 = 0;
						if (check_btn_varchange(BTN_SECONDS_DN, &pulse_length[0], 0, 99, FALSE, 0, TRUE)) my_timer_2 = 0;
					}

					//	Zifferngeblinke aus
					for (unsigned char i = 0; i < 3; i++) blink_digit[i] = FALSE;
				}


				//	GGf. geänderte Einstellungen im EEPROM speichern
				if (brightness_read != brightness) { eeprom_write_byte(BRIGHTNESS_EEPROM_ADDR_A, brightness); brightness_read = brightness; }
				if (pulse_length_read[0] != pulse_length[0]) { eeprom_write_byte(PULSE_LENGTH_EEPROM_ADDR_A, pulse_length[0]); pulse_length_read[0] = pulse_length[0]; }
				if (pulse_length_read[1] != pulse_length[1]) { eeprom_write_byte(PULSE_LENGTH_EEPROM_ADDR_B, pulse_length[1]); pulse_length_read[1] = pulse_length[1]; }
				if (pulse_length_read[2] != pulse_length[2]) { eeprom_write_byte(PULSE_LENGTH_EEPROM_ADDR_C, pulse_length[2]); pulse_length_read[2] = pulse_length[2]; }
				if (shots_total_read != shots_total) { eeprom_write_byte(SHOTS_TOTAL_EEPROM_ADDR_A, shots_total); shots_total_read = shots_total; }
			}

			//	Zeit gesetzt
			if (timer_set)
			{

				//	Zeit auf 00:00:00: Testfunktion
				if (timer_max[HOURS] == 0 && timer_max[MINUTES] == 0 && timer_max[SECONDS] == 0)
				{
					//	Timer stoppt
					timer_running = FALSE;

					//	Alle Segmente des Displays leuchten auf
					for (unsigned char i = 0; i < 6; i++)
					{
						display[i] = 8;
						dot[i] = 1;
					}

					//	Relais zieht verzögert an
					my_timer = 0; while (my_timer < 100);
					set_bit(RELAIS_port, RELAIS_bit);

					//	Warten, bis Schalter wieder auf SETUP steht
					while (!btn_press(BTN_SETUP, FALSE));

					//	Relais schaltet verzögert wieder ab
					my_timer = 0; while (my_timer < 100);
					clear_bit(RELAIS_port, RELAIS_bit);

					//	Punkte wieder ausschalten
					for (unsigned char i = 0; i < 6; i++) dot[i] = 0;
				}
				else
				{
					//	Timer läuft
					if (!timer_running)
					{
						timer_running = TRUE;

						//	Timer laden
						timer_run[HOURS] = timer_max[HOURS];
						timer_run[MINUTES] = timer_max[MINUTES];
						timer_run[SECONDS] = timer_max[SECONDS];
						minutes_max = 59;

						//	Bei Timerstart Millisekundenzähler löschen
						mseconds = 0;
					}

					//	Stunden, Minuten, Sekunden durch zwei Punkte trennen
					dot[2] = 1;
					dot[4] = 1;

					//	Timer erreicht 0, neu laden
					if (timer_run[HOURS] == 0 && timer_run[MINUTES] == 0 && timer_run[SECONDS] == 0)
					{
						//	Relais ein
						relais = TRUE;
						dot[0] = 1;
						set_bit(RELAIS_port, RELAIS_bit);

						if ((pulse_length[0] + pulse_length[1] + pulse_length[2]) > 0)
						{
							//	Timer mit Pulslänge laden
							timer_run[HOURS] = 0;
							timer_run[MINUTES] = pulse_length[2];
							timer_run[SECONDS] = pulse_length[1];
							minutes_max = 99;

							//	Mittlere beide Punkte blinken lassen
							dot_blink_middle = TRUE;

							//	Millisekundenzähler resetten
							mseconds = 0;

							//	Minuten, Sekunden abwarten bzw. bei gedrücktem Setup-Taster abbrechen
							while (((timer_run[MINUTES] + timer_run[SECONDS]) > 0) && !btn_press(BTN_SETUP, FALSE));

							//	Millisekunden abwarten bzw. bei gedrücktem Setup-Taster abbrechen
							while ((mseconds < (pulse_length[0] * 10)) && !btn_press(BTN_SETUP, FALSE));

							//	Millisekundenzähler resetten
							mseconds = 0;

							//	Relais aus
							relais = FALSE;
							dot[0] = 0;
							clear_bit(RELAIS_port, RELAIS_bit);

							//	Mittlere beide Punkte nicht mehr blinken lassen
							dot_blink_middle = FALSE;
						}

						//	Timer laden
						timer_run[HOURS] = timer_max[HOURS];
						timer_run[MINUTES] = timer_max[MINUTES];
						timer_run[SECONDS] = timer_max[SECONDS];
						minutes_max = 59;

						//	Wenn shots_total == 0, unendlich Auslösungen
						if (shots_total > 0)
						{
							if (shots_total_this + 1 < shots_total)
							{
							//	Schusszähler inkrementieren
								shots_total_this++;
							}
							else
							{
								//	Zum Parameter-Setup springen
								timer_running = FALSE;
								timer_set = FALSE;

								//	Timer auf Null
								timer_max[HOURS] = 0;
								timer_max[MINUTES] = 0;
								timer_max[SECONDS] = 0;
							}
						}
					}
					else
					{
						//	Zeit anzeigen
						display_number(timer_run[SECONDS], 0);
						display_number(timer_run[MINUTES], 1);
						display_number(timer_run[HOURS], 2);
					}
				}
			}
		}

		//	Relais nach vordefinierter Zeit ggf. wieder ausschalten (im pulse_length == 0 Modus)
		if (relais && mseconds >= RELAIS_PULSE_LENGTH)
		{
			relais = FALSE;
			dot[0] = 0;
			clear_bit(RELAIS_port, RELAIS_bit);
		}
	}
}


//	Initialisierung
void init()
{
	//	PORTA
	DDRA =  0b11111111;
	PORTA = 0b00000000;

	//	PORTB
	DDRB =  0b00000001;
	PORTB = 0b00000000;

	//	PORTC
	DDRC =  0b00111111;
	PORTC = 0b00000000;

	//	PORTD
	DDRD =  0b00000000;
	PORTD = 0b11111111;

	//	Interrupts
	cli();

	//	TIMER1
	TCCR1B = (1 << WGM12) | (1 << CS10);	//	CTC-Mode, Prescaler 1
	OCR1AL = 200;
	OCR1AH = 0;
	set_bit(TIMSK, OCIE1A);	//	Output Compare Match Interrupt Enable

	//	TIMER2
	TCCR2 = (1 << WGM21) | 0b010;	//	CTC-Mode, Prescaler 8
	OCR2 = 249;
	set_bit(TIMSK, OCIE2);	//	Output Compare Match Interrupt Enable

	sei();
}


//	TIMER1: Display-Multiplexing, -PWM
ISR(TIMER1_COMPA_vect)
{
	//	Multiplexingmodus und PWM-Vergleichswert setzen
	brightness_low = brightness_low_table[brightness - 1];
	pwm_compare = pwm_compare_table[brightness - 1];

	static unsigned char display_num = 0;

	//	Unschönes Konstrukt, aber was will man machen xD
	if (brightness_low)
	{
		//	Alle Anzeigen auf einmal multiplexen -> Dunkles Display
		for (display_num = 0; display_num < 6; display_num++)
		{
			//	Aktive Kathode aus
			CATHODES_port = 0;

			if (pwm_counter < pwm_compare)
			{
				//	Anoden für die aktive Anzeige setzen
				if (display[display_num] != -1 && !(blink_digit[(int)(display_num / 2)] && mseconds >= 499))
				{
					ANODES_port = anodes_table[(unsigned char)display[display_num]];
				}
				else
				{
					ANODES_port = 0;
				}

				//	Punkt ggf. blinken lassen
				if (dot[display_num] && !(mseconds >= 499 && (blink || (dot_blink_middle && (display_num == 2 || display_num == 4))))) ANODES_port |= dot[display_num] * SEG_DP;

				//	Aktive Kathode an, Anzeigen von rechts nach links durchlaufen
				CATHODES_port = 1 << display_num;
			}
		}

		//	Aktive Kathode aus
		CATHODES_port = 0;
	}
	else
	{
		//	Nur eine Anzeige pro Interruptaufruf multiplexen -> Helles Display

		//	Aktive Kathode aus
		CATHODES_port = 0;

		if (pwm_counter < pwm_compare)
		{
			//	Anoden für die aktive Anzeige setzen
			if (display[display_num] != -1 && !(blink_digit[(int)(display_num / 2)] && mseconds >= 499))
			{
				ANODES_port = anodes_table[(unsigned char)display[display_num]];
			}
			else
			{
				ANODES_port = 0;
			}

			//	Punkt ggf. blinken lassen
			if (dot[display_num] && !(mseconds >= 499 && (blink || (dot_blink_middle && (display_num == 2 || display_num == 4))))) ANODES_port |= dot[display_num] * SEG_DP;

			//	Aktive Kathode an, Anzeigen von rechts nach links durchlaufen
			CATHODES_port = 1 << display_num;
			if (display_num < 5) display_num++; else display_num = 0;
		}
	}

	if (pwm_counter < PWM_COUNTER_MAX) pwm_counter++; else pwm_counter = 0;
}


//	TIMER2: Millisekundenzähler, Timer, Tastenzähler
ISR(TIMER2_COMP_vect)
{
	//	Benutzerdefinierte Timer
	my_timer++;
	my_timer_2++;

	if (mseconds < 999)
	{
		//	Millisekunden zählen
		mseconds++;
	}
	else
	{
		mseconds = 0;

		//	Ggf. laufenden Timer dekrementieren
		if (timer_running)
		{
			if (timer_run[SECONDS] > 0)
			{
				timer_run[SECONDS]--;
			}
			else
			{
				timer_run[SECONDS] = 59;
				if (timer_run[MINUTES] > 0)
				{
					timer_run[MINUTES]--;
				}
				else
				{
					timer_run[MINUTES] = minutes_max;
					if (timer_run[HOURS] > 0)
					{
						timer_run[HOURS]--;
					}
					else
					{
						timer_run[HOURS] = 99;
					}
				}
			}
		}
	}

	//	Tastenzähler

	//	Start-/Setupschalter (Ein/Ein)
	if (bit_is_clear(BTN_START_pin, BTN_START_bit) && button[BTN_START] < BTN_COUNTER_MAX)
	{
		button[BTN_START]++;
		button[BTN_SETUP] = 0;
	}
	if (bit_is_clear(BTN_SETUP_pin, BTN_SETUP_bit) && button[BTN_SETUP] < BTN_COUNTER_MAX)
	{
		button[BTN_SETUP]++;
		button[BTN_START] = 0;
	}

	//	Stundentaster (Ein/Aus/Ein)
	if (bit_is_clear(BTN_HOURS_UP_pin, BTN_HOURS_UP_bit) && button[BTN_HOURS_UP] < BTN_COUNTER_MAX) { button[BTN_HOURS_UP]++; }
	if (bit_is_clear(BTN_HOURS_DN_pin, BTN_HOURS_DN_bit) && button[BTN_HOURS_DN] < BTN_COUNTER_MAX) { button[BTN_HOURS_DN]++; }
	if (bit_is_set(BTN_HOURS_UP_pin, BTN_HOURS_UP_bit) && bit_is_set(BTN_HOURS_DN_pin, BTN_HOURS_DN_bit))
	{
		button[BTN_HOURS_UP] = 0;
		button[BTN_HOURS_DN] = 0;
	}

	//	Minutentaster (Ein/Aus/Ein)
	if (bit_is_clear(BTN_MINUTES_UP_pin, BTN_MINUTES_UP_bit) && button[BTN_MINUTES_UP] < BTN_COUNTER_MAX) button[BTN_MINUTES_UP]++;
	if (bit_is_clear(BTN_MINUTES_DN_pin, BTN_MINUTES_DN_bit) && button[BTN_MINUTES_DN] < BTN_COUNTER_MAX) button[BTN_MINUTES_DN]++;
	if (bit_is_set(BTN_MINUTES_UP_pin, BTN_MINUTES_UP_bit) && bit_is_set(BTN_MINUTES_DN_pin, BTN_MINUTES_DN_bit))
	{
		button[BTN_MINUTES_UP] = 0;
		button[BTN_MINUTES_DN] = 0;
	}

	//	Sekundentaster (Ein/Aus/Ein)
	if (bit_is_clear(BTN_SECONDS_UP_pin, BTN_SECONDS_UP_bit) && button[BTN_SECONDS_UP] < BTN_COUNTER_MAX) button[BTN_SECONDS_UP]++;
	if (bit_is_clear(BTN_SECONDS_DN_pin, BTN_SECONDS_DN_bit) && button[BTN_SECONDS_DN] < BTN_COUNTER_MAX) button[BTN_SECONDS_DN]++;
	if (bit_is_set(BTN_SECONDS_UP_pin, BTN_SECONDS_UP_bit) && bit_is_set(BTN_SECONDS_DN_pin, BTN_SECONDS_DN_bit))
	{
		button[BTN_SECONDS_UP] = 0;
		button[BTN_SECONDS_DN] = 0;
	}
}


//	Zahl in zwei Ziffern zerlegen und auf eine Displaydoppelstelle schreiben
void display_number(unsigned char number, unsigned char d)
{
	//	Zahlen in einzelne Ziffern umwandeln
	unsigned char digit_10 = number / 10;
	unsigned char digit_1 = number - digit_10 * 10;
	display[d * 2] = digit_1;
	display[d * 2 + 1] = digit_10;
}


//	Knopfdruck erfassen
char btn_press(unsigned char btn, char reset)
{
	if (button[btn] >= BTN_PRESS_THRESHOLD)
	{
		if (reset) button[btn] = 0;
		return TRUE;
	}
	else
	{
		return FALSE;
	}
}


//	Gehaltenen Knopfdruck erfassen
char btn_hold(unsigned char btn)
{
	return (button[btn] >= BTN_HOLD_THRESHOLD);
}


//	Alle Zeiteinstellungs-Tastenzähler bis auf den angegebenen löschen
void btn_reset_all_except(unsigned char btn)
{
	if (btn != BTN_HOURS_UP) button[BTN_HOURS_UP] = 0;
	if (btn != BTN_HOURS_DN) button[BTN_HOURS_DN] = 0;
	if (btn != BTN_MINUTES_UP) button[BTN_MINUTES_UP] = 0;
	if (btn != BTN_MINUTES_DN) button[BTN_MINUTES_DN] = 0;
	if (btn != BTN_SECONDS_UP) button[BTN_SECONDS_UP] = 0;
	if (btn != BTN_SECONDS_DN) button[BTN_SECONDS_DN] = 0;
}


//	Überprüft, ob der Button btn gedrückt bzw. gehalten wird, inkrementiert bzw. dekrementiert die
//	Variable var im angegebenen Bereich min..max und zeigt sie auf dem Display an der Doppelstelle d an
char check_btn_varchange(unsigned char btn, volatile unsigned char *var, unsigned char min, unsigned char max, char do_inc, unsigned char d, unsigned char blink_other)
{
	if (btn_press(btn, FALSE))
	{
		if (blink_other)
		{
			blink_digit[d] = FALSE;
			//	Millisekundenzähler resetten, damit das Ziffernblinken neu startet
			mseconds = 0;
		}

		btn_reset_all_except(btn);

		if (do_inc && *var < max) (*var)++;
		if (!do_inc && *var > min) (*var)--;
		display_number(*var, d);

		unsigned int btn_hold_delay_this = BTN_HOLD_DELAY;

		while (btn_press(btn, FALSE))
		{
			//	Taster gehalten
			if (btn_hold(btn))
			{
				if ((do_inc && *var < max) || (!do_inc && *var > min))
				{
					if (do_inc && *var < max) (*var)++;
					if (!do_inc && *var > min) (*var)--;
					display_number(*var, d);
					my_timer = 0;
					while (my_timer < btn_hold_delay_this);
					if (btn_hold_delay_this >= BTN_HOLD_DELAY_MIN) btn_hold_delay_this -= BTN_HOLD_DELAY_DEC;
				}
			}
		}

		if (blink_other) blink_digit[d] = TRUE;

		return TRUE;
	}
	else
	{
		return FALSE;
	}
}
