//	Kameratimer 1.0
//
//	Copyright (C) 2009 Paul Wilhelm <paul@mosfetkiller.de>
//	http://mosfetkiller.de/?s=kameratimer
//
//	Modul: main.h | Hauptprogramm (Header)

#ifndef MAIN_H_INCLUDED
#define MAIN_H_INCLUDED

//	Externer Quarz 2 MHz
#define F_CPU	2000000UL

//	Includes
#include <util/delay.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <avr/io.h>

//	Bool
#define FALSE 0
#define TRUE 1

//	Bitpopelei
#define	set_bit(var, bit)	((var) |= (1 << (bit)))
#define	clear_bit(var, bit)	((var) &= (unsigned)~(1 << (bit)))

//	Pinout
#define ANODES_port		PORTA
#define CATHODES_port	PORTC

#define RELAIS_port	PORTB
#define RELAIS_bit	0

//	Anoden
#define SEG_A	(1 << 0)
#define SEG_B	(1 << 1)
#define SEG_C	(1 << 2)
#define SEG_D	(1 << 3)
#define SEG_E	(1 << 4)
#define SEG_F	(1 << 5)
#define SEG_G	(1 << 6)
#define SEG_DP	(1 << 7)

//	Schalter/Taster
#define BTN_START			0
#define BTN_START_pin		PIND
#define BTN_START_bit		7

#define BTN_SETUP			1
#define BTN_SETUP_pin		PIND
#define BTN_SETUP_bit		6

#define BTN_HOURS_UP		2
#define BTN_HOURS_UP_pin	PIND
#define BTN_HOURS_UP_bit	5

#define BTN_HOURS_DN		3
#define BTN_HOURS_DN_pin	PIND
#define BTN_HOURS_DN_bit	4

#define BTN_MINUTES_UP		4
#define BTN_MINUTES_UP_pin	PIND
#define BTN_MINUTES_UP_bit	3

#define BTN_MINUTES_DN		5
#define BTN_MINUTES_DN_pin	PIND
#define BTN_MINUTES_DN_bit	2

#define BTN_SECONDS_UP		6
#define BTN_SECONDS_UP_pin	PIND
#define BTN_SECONDS_UP_bit	1

#define BTN_SECONDS_DN		7
#define BTN_SECONDS_DN_pin	PIND
#define BTN_SECONDS_DN_bit	0

#define BTN_COUNTER_MAX		65535
#define BTN_PRESS_THRESHOLD	20		//	ms
#define BTN_HOLD_THRESHOLD	800		//	ms
#define BTN_HOLD_DELAY		400		//	ms
#define BTN_HOLD_DELAY_DEC	20		//	ms
#define BTN_HOLD_DELAY_MIN	60		//	ms

//	pulse_length == 0
#define RELAIS_PULSE_LENGTH	100		//	ms (max. 999)

#define PULSE_LENGTH_MENU_IDLE_EXIT 3000	//	ms (max. 65535)

#define PWM_COUNTER_MAX	99

#define BRIGHTNESS_EEPROM_ADDR_A	(unsigned char*)0
#define PULSE_LENGTH_EEPROM_ADDR_A	(unsigned char*)1
#define PULSE_LENGTH_EEPROM_ADDR_B	(unsigned char*)2
#define PULSE_LENGTH_EEPROM_ADDR_C	(unsigned char*)3
#define SHOTS_TOTAL_EEPROM_ADDR_A	(unsigned char*)4

//	Timer
#define SECONDS	0
#define MINUTES	1
#define HOURS	2

//	Variablen
volatile char display[6] = { -1, -1, -1, -1, -1, -1 };
volatile unsigned char dot[6], blink = FALSE, blink_digit[3] = { FALSE, FALSE, FALSE }, brightness_low = FALSE, dot_blink_middle = FALSE;
volatile unsigned int pwm_counter, pwm_compare = PWM_COUNTER_MAX;
volatile unsigned char brightness = 10;

char brightness_low_table[10] = { TRUE, TRUE, TRUE, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE };
unsigned char pwm_compare_table[10] = { 5, 10, 20, 30, 99, 15, 30, 50, 70, 99 };

volatile unsigned int mseconds = 0;
volatile unsigned int button[8];

volatile unsigned int my_timer = 0, my_timer_2 = 0;

volatile unsigned char timer_run[3], timer_max[3], timer_set = FALSE, timer_running = FALSE, relais = FALSE, minutes_max = 59;

volatile unsigned char pulse_length[3], shots_total = 0;
unsigned char shots_total_this;

//	Anodentabelle
unsigned char anodes_table[10] =
{
	(SEG_A | SEG_B | SEG_C | SEG_D | SEG_E | SEG_F),			//	0
	(SEG_B | SEG_C),											//	1
	(SEG_A | SEG_B | SEG_D | SEG_E | SEG_D | SEG_G),			//	2
	(SEG_A | SEG_B | SEG_C | SEG_D | SEG_G),					//	3
	(SEG_B | SEG_C | SEG_F | SEG_G),							//	4
	(SEG_A | SEG_C | SEG_D | SEG_F | SEG_G),					//	5
	(SEG_A | SEG_C | SEG_D | SEG_E | SEG_F | SEG_G),			//	6
	(SEG_A | SEG_B | SEG_C),									//	7
	(SEG_A | SEG_B | SEG_C | SEG_D | SEG_E | SEG_F | SEG_G),	//	8
	(SEG_A | SEG_B | SEG_C | SEG_D | SEG_F | SEG_G)				//	9
};

//	Prototypen
void init();
void display_number(unsigned char number, unsigned char d);
char btn_press(unsigned char btn, char reset);
char btn_hold(unsigned char btn);
void btn_reset_all_except(unsigned char btn);
char check_btn_varchange(unsigned char btn, volatile unsigned char *var, unsigned char min, unsigned char max, char do_inc, unsigned char d, unsigned char blink_other);

#endif
